function params = mix(X1, Init, iters, verbose, X2, S, implicit_noisevar)

% Run the iteration formulae for the model parameters without mex-functions.
% 
% INPUT
% Init  The input model params struct
% iters The amount of iterations run
% X1    The real-valued data
% X2    The nominal-valued data
% S     The amount of categories/possible values in the nominals
% implicit_noisevar   Variance of implicit noise
%
% OUTPUT
% new model parameter struct
%
% author Jeremias Seppa


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.


X1=X1';   %for compliance with stupid legacy aspects of mex modeller

%nonmex=1  %%%%%%%%%!!!!!!
global IVGA_CONFIG;
if ~isfield( IVGA_CONFIG, 'MDL_VQ_RN_NONMEX_IN_USE' )
  IVGA_CONFIG.MDL_VQ_RN_NONMEX_IN_USE = true;
  if ~IVGA_CONFIG.QUIET
    fprintf(['\n\n*** The C implementation of the Gaussian mixture model ' ...
             'is not compiled for\n    this platform! The slower Matlab ' ...
             'implementation is used instead. You can\n    compile the C ' ...
             'implementation for the current platform with the ' ...
             'command\n    ''ivga_compile''.\n' ...
             '    See the README file for more details.\n\n']);
  end
  warning( 'Failed to use the C implementation of the modeler!' );
end

Uhat=Init.U_hat;
U_p=Init.U_p;
Mu_bar=Init.Mu_bar';
Mu_tilde=Init.Mu_tilde';
Mu_mu=Init.Mu_mu;
S2_mu=Init.S2_mu;
Alpha_ksi=Init.Alpha_ksi;
Beta_ksi=Init.Beta_ksi;
Ksi_alpha=Init.Ksi_alpha';
Ksi_beta=Init.Ksi_beta';
c_p=Init.C_prior;
logc=Init.logC;

[N,M1]=size(X1);
M2=size(X2,2);
C=size(Mu_bar,1);

Ksi_log=psi(Ksi_alpha)-log(Ksi_beta);  % a not-too-much-helping
                                       % helper variable

for iter=1:iters,	   
  temp=zeros(C,N);
  if M2
    for j=1:M2,
	UhatPsi=psi(Uhat{j});
        temp=temp-UhatPsi(:,X2(:,j))+repmat(psi(sum(Uhat{j},2)),1,N);
    end
  end 
  if M1
    S2_x=Ksi_beta./Ksi_alpha;
    S2_x(S2_x < 1e-100) = 1e-100;
    for j=1:M1
      temp=temp+0.5*((repmat(Mu_tilde(:,j),1,N)+(repmat(X1(:,j)',C,1)- ...
                   repmat(Mu_bar(:,j),1,N)).^2 + implicit_noisevar)./repmat(S2_x(:,j),1,N) ...
          - repmat(Ksi_log(:,j),1,N));
    end
  end
  W=(temp-repmat(logc,1,N));   
  W=exp(repmat(min(W,[],1),C,1)-W);   
  W=W./repmat(sum(W,1),C,1);         % 'normalise'
  F=sum(W,2);
  logc=psi(F+c_p)-psi((C*c_p)+sum(F));
  %% Update the mixture components, i.e. "centroids" :
  if M2
    for j=1:M2,
      for k=1:S(j),
        Uhat{j}(:,k)= U_p(j) + W*(X2(:,j)==k); % matrix prod. :) 
      end
    end
  end
  if M1
    S2_muM=repmat(S2_mu,C,1);
    FM=repmat(F,1,M1);
    Mu_bar = (S2_x .* repmat(Mu_mu,C,1) + S2_muM .* ...
	      (W*X1))./(S2_x + S2_muM.*FM);
    Mu_tilde = S2_x.*S2_muM./(S2_x + S2_muM.*FM);
    Ksi_alpha = repmat(Alpha_ksi,C,1)+0.5*FM;
    for j=1:M1
      Ksi_beta(:,j)=Beta_ksi(j) + 0.5 * ...
	  sum(W.*(repmat(Mu_tilde(:,j),1,N)+(repmat(X1(:,j)',C,1)- ...
		repmat(Mu_bar(:,j),1,N)).^2 + implicit_noisevar),2);
    end
    % Update some priors .. (and calculate Ksi_Log) :
    Mu_mu=(1/C)*sum(Mu_bar,1);
    S2_mu=(1/C)*sum(Mu_tilde + ((Mu_bar-repmat(Mu_mu,C,1)).^2),1);
    T=repmat(C,1,M1)./sum(Ksi_alpha./Ksi_beta,1);
    Ksi_log=psi(Ksi_alpha)-log(Ksi_beta);
    Alpha_ksi=Alpha_ksi+repmat(0.5,1,M1)./(psi(Alpha_ksi)-log(Alpha_ksi)) + ...
	  	 repmat(0.5,1,M1)./(repmat(1/C,1,M1).*sum(-Ksi_log,1) - log(T));
  
    Beta_ksi=Alpha_ksi.*T;
  end
end

params.U_hat = Uhat;
params.U_p = U_p;
params.Mu_bar = Mu_bar';
params.Mu_tilde = Mu_tilde';
params.Mu_mu = Mu_mu;
params.S2_mu = S2_mu;
params.Alpha_ksi = Alpha_ksi;
params.Beta_ksi = Beta_ksi;
params.Ksi_alpha = Ksi_alpha';
params.Ksi_beta = Ksi_beta';
params.C_prior = c_p;
params.logC = logc;
params.F=F;
params.W=W;
return;







