function Params = vectquantM(X, types, S, Init, max_iters, verbose, implicit_noisevar)

% Params = vectquantM(X, types, S,[ncentroids seed], max_iters, verbose)
% Params = vectquantM(X, types, S,Init, max_iters, verbose)
%
% INPUT ARGUMENTS
%
%
%   ncentroids   number of codebook vectors
%   seed         seed number for random number generator
%   Init         struct consisting of model parameters, for example 
%                output argument Params from some previous run
%   [max_iters]  maximum number of iterations, default = 100
%                if max_iters == 1, it will be silently raised to 2.   (*
%   verbose      currently has no effect.
%
% OUTPUT ARGUMENTS
%  
%   Params       struct consisting of model parameters
%


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

X1=X(:,(types==1));
X2=X(:,(types==2));
S=S(types==2);
[n dim1]        = size(X1);
dim2= size(X2,2);

old_rand_state = [];
if max_iters == 1, max_iters = 2; end

%
% initialize
%

if isstruct(Init)
     
  % continue earlier run

else
  % make a new vq
  
  ncentroids    = Init(1);
  seed          = Init(2);
  
  Mean          = sum(X1, 1)/n;
  Var           = sum((X1-repmat(Mean,n,1)).^2, 1)/n; % variance
  
  % priors for distribution of codebook vectors Mu ~ N(Mu_Mu, S2_Mu)..

  Init          = struct('Mu_mu', Mean, 'S2_mu', Var);
  if dim2
    for j=1:dim2
      Init.U_p(j)      = 0.5;     %%
    end
  else
    Init.U_p=Inf;
  end
  % posteriors for distribution of codebook vectors Mu ~ N(Mu_bar, Mu_tilde)..

  old_rand_state = rand('state');
  rand('state', seed);
  Inds          = randperm(n);
  Init.Mu_bar   = X1(Inds(1:ncentroids), :);
  Init.Mu_tilde = repmat(1e-100,ncentroids,dim1);
  Uhat={};
  for j=1:dim2
    Uhat{j}=0.7+0.5*rand(ncentroids,S(j));      %%
    for i=1:ncentroids
      k=X2(Inds(i),j);
      Uhat{j}(i,k)=Uhat{j}(i,k)+3;              %%
    end
  end  
  Init.U_hat    = Uhat;
  
  % priors for data variance Ksi ~ Gamma(Alpha_ksi, Beta_ksi)
  
  Init.Alpha_ksi = ones(1, dim1);
  Init.Beta_ksi  = Var/ncentroids;

  % posteriors for data variance Ksi ~ Gamma(Ksi_alpha, Ksi_beta)

  Init.Ksi_alpha = ones(ncentroids, dim1);
  Init.Ksi_beta  = repmat((Var/ncentroids),ncentroids,1);
  
  % winners for data points C ~ Dirichlet(F) (F contains winner frequencies)
  
  Init.C_prior  = 0.5;
  Init.logC     = log(ones(ncentroids,1)/ncentroids);
end

Init.Mu_bar=Init.Mu_bar';
Init.Mu_tilde=Init.Mu_tilde';
Init.Ksi_alpha=Init.Ksi_alpha';
Init.Ksi_beta=Init.Ksi_beta';

Params = mix(X1', Init, max_iters, verbose, X2, S, implicit_noisevar);

Params.Mu_bar=Params.Mu_bar';
Params.Mu_tilde=Params.Mu_tilde';
Params.Ksi_alpha=Params.Ksi_alpha';
Params.Ksi_beta=Params.Ksi_beta';

if ~isempty(old_rand_state), rand( 'state', old_rand_state ); end

return;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%





