function this = measure( this, op_name, mode, State, ex_vars )

% measure( this, op_name, mode, State, [ex_vars] )
%
% op_name    name of the measured operation
% mode       'start', 'checkpoint' or 'stop'
% State      current grouping
% ex_vars    extra variables as a horizontal vector. these are stored on
%            each checkpoint without any processing. 
%            Note: If recording values that always stay constant over the
%            call, it might be a good idea to record a zero on
%            measure('start'), so that the absolute value can be seen also
%            at call level. (ok, zero iterations and the value is lost..)
%
% Measures execution time, total cost, efficiency etc.. Does nothing if the
% stats object is in inactive state.
%
% Note: if on some iteration the cputimer delta is zero but cost delta is
% not, then a zero efficiency is recorded instead of a NaN! (see method
% comments for the reason)
%
% modes:
%   start        starts recording a single iteration loop for the operation.
%                also executes once in checkpoint mode.
%   checkpoint   stores state data into the current iteration loop.
%   stop         ends recording the operation's iteration loop.
%
% example:
%   s = stats;
%   s = set_active( s, 1 );
%   s = cputimer( s, 'start' );
%   s = add_operation( s, 'improve' );
%   s = measure( s, 'improve', 'start', State );
%   while ...
%     State = improve( State );
%     s = measure( s, 'improve', 'checkpoint', State );
%   end
%   s = measure( s, 'improve', 'stop', State );


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.



% Starting state of an ongoing operation is temporarily stored into cell
% array 'startstates', which contains an entry for each operation type.
% These entries act as stacks containing the starting states of individual
% activations, most recent on the top.
%
% Efficiency is _always_ NaN on op start row and non-NaN on iteration rows!
% This variable is used in flt_stats_reader when iteration counting is
% requested. This causes one problem: if cputime delta is zero but cost
% delta is not, then a NaN efficiency cannot be recorded, because it is
% already reserved to signal 'no input' in filter chains (instead a zero is
% recorded, inf would be disastrous).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if ~this.active, return; end


if nargin < 5, ex_vars = []; end
op_ind = get_op_ind( this, op_name );


switch mode
  
 case 'start'
  
  % The operation is starting, store info from current state and run also
  % the checkpoint case.
  
  % Process the current state and push results to state stack
  this.startstates{op_ind} = ...
      { op_start( this, this.statevars, State ), this.startstates{op_ind} };
  
  % indent and reserve space if print requested
  if ~isempty( this.print_vars{op_ind} ) && ...
        ~this.statevars.internal_quiet
    this.print_indent = this.print_indent + 8;
    if this.statevars.internal_addr_cursor
      fprintf( repmat( '\n', 1, 1 + length(this.print_vars{op_ind}) ) );
    end
  end
  
  % Run in checkpoint mode
  this = measure( this, op_name, 'checkpoint', State, ex_vars );
  
 case 'checkpoint'
  
  % Store progress information into the corresponding data matrix and update
  % the start state.
  
  % get progress information and update the start state
  [datarow, this.startstates{op_ind}{1}] = ...
      op_checkpoint( this, this.statevars, State, ...
                     this.startstates{op_ind}{1}, ex_vars );
  
  % store progress information
  this.results{op_ind} = [ this.results{op_ind} ; datarow ];
  
  % print if requested
  if ~isempty( this.print_vars{op_ind} ) && ...
        ~this.statevars.internal_quiet
    indent_str = repmat( ' ', 1, this.print_indent );
    if this.statevars.internal_addr_cursor
      fprintf( '%c[%dA', 27, 1 + length(this.print_vars{op_ind}) );
      fprintf( '%sOperation %s:%c[K\n', indent_str, op_name, 27 );
      for var=this.print_vars{op_ind}
        fprintf( '%s  %s%s %10g%c[K\n', ...
                 indent_str, this.var_names{op_ind}{var}, ...
                 repmat( '.', 1, ...
                         34 - length(this.var_names{op_ind}{var}) ), ...
                 datarow(var), 27 );
      end
    else
      if cputimer(this) - this.statevars.internal_last_print > 10
        this.statevars.internal_last_print = cputimer(this);
        fprintf( '%sOperation %s:\n', indent_str, op_name );
        for var=this.print_vars{op_ind}
          fprintf( '%s  %s%s %10g\n', ...
                   indent_str, this.var_names{op_ind}{var}, ...
                   repmat( '.', 1, ...
                           34 - length(this.var_names{op_ind}{var}) ), ...
                   datarow(var) );
        end
      end
    end
  end
  
 case 'stop'
  
  % Mark the last operation's area in the corresponding result matrix and pop
  % the start state from state stack
  
  % mark the filled result matrix area
  this.areas{op_ind} = ...
      [ this.areas{op_ind} ;
        size( this.results{op_ind}, 1 ) + 1 ];
  
  % pop start state from stack
  this.startstates{op_ind} = this.startstates{op_ind}{2};
  
  % unindent and 'release' space if print requested
  if ~isempty( this.print_vars{op_ind} ) && ...
        ~this.statevars.internal_quiet
    this.print_indent = this.print_indent - 8;
    if this.statevars.internal_addr_cursor
      fprintf( repmat( [27 '[1A' 27 '[K'], 1, ...
                       1 + length(this.print_vars{op_ind}) ) );
    end
  end
  
end


return;








function start_state = op_start( this, statevars, State )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% this assumes that the checkpoint case is run immediately after this
start_state.prev_cputimer = NaN;
start_state.prev_cost     = NaN;

return;




function [datarow, start_state] = op_checkpoint( this, statevars, State, ...
                                                 start_state, ex_vars )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% get current cputimer and cost readings
mycputimer = cputimer(this);
cost       = total_cost( State );

d_cost = cost - start_state.prev_cost;
d_time = mycputimer - start_state.prev_cputimer;

if d_time == 0
  lap_efficiency = 0;
else
  lap_efficiency = - d_cost / d_time;
end


datarow = [ mycputimer, cost, lap_efficiency, ex_vars ];

start_state.prev_cputimer = mycputimer;
start_state.prev_cost     = cost;


return;
