function res = environment( query, varargin )

% result = environment( query, ... )
%
% This function returns various information about the underlying system, or
% executes system-dependent operations. When using the program for the first
% time in a new environment, this file should be first checked and modified as
% needed.
%
% Currently the query can be one of the following:
%
%   result = environment( 'has_feature', name )
%   
%     This returns a boolean value indicating the presence of the requested
%     feature. The feature names and values are defined in this file in the
%     helper function has_feature().
%
%   result = environment( 'get_system_load' )
%
%     This returns the current average system load (the maximum of averages
%     over the last 1, 5 and 15 minutes). If this information cannot be
%     provided, then NaN is returned.


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

res = feval( query, varargin{:} );

return;




function res = has_feature( name )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

switch name
  
 case 'sleep'
  % sleep: does the underlying shell support sleeping for n seconds with the
  % command: sleep n
  
  global IVGA_CONFIG;
  if ~isfield( IVGA_CONFIG, 'ENVIRONMENT_SLEEP_FEATURE' )
    t0 = clock;
    status = system('sleep 1');
    dt = etime( clock, t0 );
    IVGA_CONFIG.ENVIRONMENT_SLEEP_FEATURE = ...
        status == 0 && dt > 0.99;
  end
  
  res = IVGA_CONFIG.ENVIRONMENT_SLEEP_FEATURE;
  
 case 'addressable_cursor'
  % addressable_cursor: does the current terminal have an addressable cursor?
  
  [retcode,val] = system( 'infocmp -L|grep -q cursor_up' );
  res = (retcode == 0);
  
 case 'working_copymovefile'
  % has this Matlab a working copyfile and movefile? (some versions have bugs
  % in these)
  
  global IVGA_CONFIG;
  if ~isfield( IVGA_CONFIG, 'ENVIRONMENT_WORKING_COPYMOVEFILE' )
    
    dummy = [];
    IVGA_CONFIG.ENVIRONMENT_WORKING_COPYMOVEFILE = true;
    save( '__ivga_toolbox_envm_tmp1__.mat', 'dummy' );
    
    if ~copyfile('__ivga_toolbox_envm_tmp1__.mat', ...
                 '__ivga_toolbox_envm_tmp2__.mat') || ...
       ~movefile('__ivga_toolbox_envm_tmp1__.mat', ...
                 '__ivga_toolbox_envm_tmp3__.mat')
      IVGA_CONFIG.ENVIRONMENT_WORKING_COPYMOVEFILE = false;
      warning(['The movefile and copyfile commands seem to be broken. ' ...
               'No backup files will be used when saving new batch results.']);
    end
    
    if exist( '__ivga_toolbox_envm_tmp1__.mat', 'file' )
      delete( '__ivga_toolbox_envm_tmp1__.mat' );
    end
    if exist( '__ivga_toolbox_envm_tmp2__.mat', 'file' )
      delete( '__ivga_toolbox_envm_tmp2__.mat' );
    end
    if exist( '__ivga_toolbox_envm_tmp3__.mat', 'file' )
      delete( '__ivga_toolbox_envm_tmp3__.mat' );
    end
  
  end
  
  res = IVGA_CONFIG.ENVIRONMENT_WORKING_COPYMOVEFILE;
  
 otherwise
  error([ 'Unknown feature id: ' name ]);
  
end

return;




function res = get_system_load()

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[status, result] = system( 'uptime | sed ''s/.*load average: //''' );
if status ~= 0, res = NaN; else res = max(str2num(result)); end

return;
