function result = ivga( data, varargin )

% IVGA Find an optimized grouping for the variables of the given data.
% 
%   result = ivga( data, [random_seed], [settings], [options ...] )
%   
%   result = ivga( D )
%   result = ivga( D, 'runtime', 3600 )
%   result = ivga( D, 12345, @cs_adaptive, 'runtime', 'fastest' )
%   result = ivga( ivga_import_data( D, 'types', 'rrnnnr' ), 'runtime', 600 )
% 
% Arguments
% 
%   data              The data. This can be a numeric 2-dimensional
%                     matrix, a cell array with numeric and text cells, a file
%                     name of an Excel sheet or a text file, etc., or a data
%                     struct generated explicitly with the ivga_import_data
%                     function. Type 'help ivga_import_data' for details about
%                     supported data formats.
%                       In the case of a numeric 2-dimensional matrix, the rows
%                     represent samples and variables are along columns. All
%                     variables are modeled as real-valued by default.
%                       In case of a cell array, a text file or an Excel sheet
%                     containing both numeric and text data, all numeric
%                     variables are modeled as real-valued and variables with
%                     text labels are modeled as nominal-valued (categorical)
%                     variables.
%                       You can also use the ivga_import_data function directly
%                     to override the defaults and specify the types of the
%                     variables manually. For more information, type:
%                       help ivga_import_data
%   [random_seed]     Random number seed to be used. This can be a single
%                     real-valued number or a 35-element MATLAB random number
%                     generator state vector. If omitted or zero is given, the
%                     seed will be generated from clock. If -1 is given, then
%                     the generator will not be initialized. See section
%                     "Repeating an experiment" below.
%                     Warning:
%                       The adaptive controller (which can be selected with
%                       @cs_adaptive as the settings) uses cputime readings
%                       during operation, so experiments ran with this
%                       controller cannot be repeated identically even with the
%                       same random seed! You can use @cs_basic (which is the
%                       default) as the settings to get identically repeatable
%                       results.
%                     Default: generated from clock
%   [settings]        Defines the operation mode and default parameters for
%                     the program. This can be a function handle (e.g.,
%                     @cs_adaptive) to select one of the included presets, or a
%                     preloaded settings struct (from the 'settings' field of a
%                     result from an earlier run, for example). Note that if a
%                     preloaded struct is given, then all rest options, except
%                     the 'runs' option, are ignored!
%                     Available presets:
%                       @cs_basic         Slower, but produces identically
%                                         repeatable results. (default)
%                       @cs_adaptive      Faster, but the results cannot be
%                                         repeated identically even with the
%                                         same random seed.
%                       @cs_adaptive_2    Similar to @cs_adaptive, but
%                                         uses different learning parameters
%                                         and usually produces somewhat
%                                         different groupings. The results
%                                         cannot be repeated identically even
%                                         with the same random seed.
%                       @cs_make_models   Computes the group models and a
%                                         cost for a grouping given with the
%                                         option 'initial_grouping' (see
%                                         below). The additional option
%                                         'quality' is available with this
%                                         preset, and it can have the values
%                                         'best', 'good', 'normal', 'faster'
%                                         and 'fastest'.
%                     Default: @cs_basic
%   [options ...]     The options are explained below in the Options section.
% 
% Result
%   
%   A struct with the following fields is returned:
%     .group_count        Number of groups in the computed grouping.
%     .grouping           The computed grouping: a cell array of index
%                         vectors, so that grouping{i} contains the indices of
%                         the variables in group i.
%     .group_models       Models for the groups.
%     .cost               The cost of the grouping.
%     .random_seed        The random seed that was used. See section
%                         "Repeating an experiment" below.
%     .settings           The settings that were used. These can be used
%                         directly as the 'settings' argument of a new IVGA
%                         run. See section "Repeating an experiment" below.
%     .internal           Internal data and runtime statistics used by some
%                         helper functions when analyzing the result.
%   The functions ivga_show() and ivga_show_stats() can be used to analyze
%   the result.
% 
% Options
%   
%   Some advanced options are omitted from this help section. See the ivga.m
%   file directly to access their documentation or type the following in
%   Matlab: more on; type ivga.m
%
%   Controlling the runtime:
%
%     If no runtime option is given, then the preset 'normal' is used (see
%     below for details).
%     
%     ['runtime', preset_name]
%       Use preset settings to determine the runtime. Available choices:
%         'fastest'   Computes only a rough estimate for the grouping. Some
%                     time-consuming operations are disabled
%                     completely. Runtime is determined directly from the
%                     dimensionality of the data set and is thus very
%                     predictable.
%         'faster'    The fastest setting which still has all features enabled
%                     and decides the runtime adaptively. Runtime is based on
%                     the convergence of the grouping.
%         'normal'    The default setting. This is a reasonable compromise
%                     between speed and quality. Runtime is based on the
%                     convergence of the grouping.
%         'slower'    Produces groupings which are very likely near a local
%                     minimum (in the sense of moving individual variables
%                     between groups). Runtimes are still quite
%                     reasonable. Runtime is based on the convergence of the
%                     grouping.
%         'slowest'   Produces groupings which are very likely at a local
%                     minimum (in the sense of moving individual variables
%                     between groups), but the runtimes might be extremely
%                     long.
%                     NOTE:
%                       Using the 'slowest' runtime preset is not recommended,
%                       and usually you will get better results for the same
%                       total runtime by using the 'runs' option to execute
%                       multiple faster runs from different random
%                       initializations and using the best obtained result.
%       Example:
%         result = ivga( D, 'runtime', 'faster' )
%     
%     ['runtime', seconds]
%       Runtime, in CPU timer seconds. Note that the requested runtime might
%       get exceeded slightly in some cases.
%       Example:
%         result = ivga( D, 'runtime', 1800 )
%     
%     ['min_runtime', seconds]
%       Minimum runtime, in CPU timer seconds.
%     
%     ['max_runtime', seconds]
%       Maximum runtime, in CPU timer seconds. Note that this limit might get
%       exceeded slightly in some cases.
%     
%     ['runtime', {'iterations', n}]
%       Evaluation is stopped after n iterations of the algorithm.
%       Example:
%         result = ivga( D, 'runtime', {'iterations', 1000} )
%     
%     ['runtime', {'variable_sweeps', n}]
%       Evaluation is stopped after attempting approximately n times every
%       possible variable move. Both successful and failed attempts are
%       counted.
%     
%     ['runtime', {'stationary_groupcount_minimum', n}]
%       Evaluation is stopped if the groupcount minimum has not decreased after
%       attempting approximately n times every possible variable move. n can be
%       real-valued.
%     
%     ['runtime', { 'stationary_variable_sweeps', n,
%                   ['sweeps_to_stabilize', s] }]
%       Evaluation is stopped if the grouping has not changed after attempting
%       n times all possible moves of individual variables. n can be
%       real-valued.
%         The continuous fine-tuning of the component models during the
%       process can keep the grouping oscillating around a local minimum for a
%       while. To prevent this from interfering with the
%       'stationary_variable_sweeps' stopping test, the stabilization feature
%       can be used to disable all fine-tuning and non-local operations for a
%       while when it seems that a local minimum has been reached: when the
%       stationary variable sweeps counter reaches s, all operations except
%       moving individual variables are disabled. If the grouping now changes
%       before reaching the stopping level n, then stabilization is disabled
%       again and normal operation is continued. s should be smaller than
%       n. Stabilization is disabled by default.
%       Examples:
%         result = ivga( D, 'runtime', {'stationary_variable_sweeps', 0.5} )
%         result = ivga( D, 'runtime', {'stationary_variable_sweeps', 1.0,
%                                       'sweeps_to_stabilize', 0.5} )
%     
%   Specifying an initial grouping:
%     
%     ['initial_grouping', g]
%       The grouping g is used as the initial grouping. g should be a cell
%       array of index vectors, so that g{i} contains the indices of the
%       variables in group i. All variables in the data set must be included in
%       the grouping. The 'allow_partial_grouping' flag should be specified
%       in addition if some variables are left out intentionally.
%       Example:
%         g = {[1 2], [3 4], 5:100}
%     
%     ['allow_partial_grouping']
%       If this flag is specified, then the grouping provided with the
%       'initial_grouping' option is allowed to not contain all of the
%       variables in the data set.
%     
%   Stabilizing the Gaussian mixture model:
%
%     ['implicit_noisevar', v]
%       The value v is used as the variance of implicit noise, which
%       is added to the data. This keeps the Gaussian mixture components
%       from collapsing to sizes smaller than that indicated by v.
%       The default value is 0, but it should be increased if the
%       Gaussian mixture code gives divide by zero warnings.
%       The same value is used for all continuous variables.
%       Suitable values might be of the order of assumed quantization
%       error in the data.
%     
%   Using the batch mode:
%     
%     ['runs', n, ['save_as', name]]
%       This option will cause n groupings to be computed from different random
%       initializations. The results are returned as a cell array of result
%       structs, sorted into ascending order by the cost of the groupings. The
%       ivga_show() and ivga_show_stats() functions accept directly also this
%       kind of result arrays. A random seed can be provided normally, and it
%       will be used to generate seeds for the individual runs.
%         The 'save_as' option causes individual results to be written to
%       disk as soon as possible, making it possible to resume an interrupted
%       run by using the 'continue' option. The final result array is also
%       stored to disk into the MATLAB file
%       'batch_<name>_results.mat'. Temporary files with the prefix
%       'batch_run_<name>_' will be created during computation and deleted when
%       no more needed. The status of a running computation can be read from
%       the file 'batch_<name>_status.txt'.
%       Example:
%         results = ivga( D, @cs_adaptive, 'runs', 10, 'save_as', 'test01' )
%         ivga_show( results{1} )
%         ivga_show_stats( results )
%         ivga_show_stats( 'test01' )
%
%     ['continue', name]
%       This can be used to continue an interrupted computation. It is
%       available only if the 'runs' and 'save_as' options were used when
%       starting the run (see above). 'name' must be the name that was given to
%       the 'save_as' option.
%       Example:
%         ivga( D, 'runs', 10, 'save_as', 'test01' )
%         <the computation is interrupted here>
%         ivga( 'continue', 'test01' )
%
%   Other:
%     
%     ['quiet']
%       All console output is suppressed.
% 
% 
% Repeating an experiment
%
% With some restrictions, an IVGA experiment can be repeated identically by
% using the information contained in an IVGA result struct. First, the original
% data set must be available. Second, an identical re-run is possible only if
% the basic controller (which is the default one) was used in the experiment.
%
% To see which controller was used to run an experiment whose result is in the
% variable 'result', give the command:
%   result.settings.ctrl.class
%
% To repeat an experiment by using its result struct 'result' (the original
% data set is in variable D):
%   new_result = ivga( D, result.random_seed, result.settings )
%
%
% More examples
%   
%   D = rand(250,2); D = [D D];
%   
%   Basic usage:
%     result = ivga( D )
%     result = ivga( D, @cs_adaptive )
%     result = ivga( D, 12345, @cs_adaptive_2 )
%   
%   Controlling the runtime:
%     result = ivga( D, 'runtime', 60 )
%     result = ivga( D, 'runtime', {'iterations', 500} )
%     result = ivga( D, 'runtime', {'iterations', 500}, 'min_runtime', 60 )
%     result = ivga( D, 'min_runtime', 60, 'max_runtime', 600 )
%     result = ivga( D, @cs_adaptive, 'runtime', 'fastest' )
%   
%   Specifying an initial grouping:
%     result = ivga( D, 'initial_grouping', {[1 2], [3 4], 5:50} )
%     result = ivga( D, 'initial_grouping', previous_result.grouping )
%   
%   Analyzing the result:
%     ivga_show( result )
%     ivga_show_stats( result )
%     ivga_show_stats( {result1, result2}, 'y_var', 'group count' )
% 
% 
% See also IVGA_IMPORT_DATA, IVGA_SHOW, IVGA_SHOW_STATS, IVGA_COMPILE


%  This file is a part of the Independent Variable Group Analysis package
% 
%  Copyright (C) 2001-2006 Esa Alhoniemi, Antti Honkela, Krista Lagus,
%  Jeremias Seppa, Harri Valpola, and Paul Wagner
% 
%  This program is free software; you can redistribute it and/or modify
%  it under the terms of the GNU General Public License as published by
%  the Free Software Foundation; either version 2, or (at your option)
%  any later version.
% 
%  This program is distributed in the hope that it will be useful,
%  but WITHOUT ANY WARRANTY; without even the implied warranty of
%  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%  GNU General Public License (included in file License.txt in the
%  program package) for more details.



% Advanced options
% ----------------
%
%   Multiple data sets in a batch run:
%   
%     When running in batch mode (activated by using the 'runs' option), the
%     'data' argument can be a one-dimensional cell array of IVGA data structs
%     or data sources, in which case the whole batch is repeated for each data
%     set (the execution order is randomized). The result will be a
%     2-dimensional array with each row containing the results for the
%     corresponding data set. The columns of a row contain the repetitions for
%     that data set. The results are still sorted within each row. The result
%     handling functions accept also these 2-dimensional results as input.
%     Example:
%       data1 = ivga_import_data( 'data1.txt', 'types', 'rrnn' )
%       result = ivga( {data1, 'data2.txt', 'data3.txt'}, 'runs', 1 )
%       result = ivga( {data1, 'data2.txt', 'data3.txt'}, 'runs', 5 )
%   
%   Specifying an initial grouping:
%     
%     ['var_percentage', n]
%       n percents of all variables are selected at random and an own group is
%       created for each. At least 2 variables are chosen always.
%     
%     ['initial_groupings', g]
%       This can be used to continue from an explicit set of groupings or from
%       a set of groupings formed by a previous batch run, so that all of the
%       groupings are tuned further in parallel. This mode can be used only
%       in batch mode.
%       The parameter g can either be:
%         a) { 'source', source, [entry] }, where the source argument defines
%            the result source as explained in ivga_show() and
%            ivga_show_stats(), and entry is the index of the entry whose final
%            groupings are to be used. The entry argument can be omitted, in
%            which case the first entry will be used.
%         b) a cell array, so that g{dataset,iter} contains a cell array n,
%            so that n{group} contains an index vector with variable indices
%            for the group 'group'.
%       NOTE: The dimensionality of the provided groupings must be equal or
%       larger than the number of data sets and repetitions in the batch to be
%       executed!
% 
%   Other:
%     
%     ['max_split_depth', depth]
%       Maximum split recursion depth. Default: from the settings file


% Internal
% --------
%
% Additional features:
%
% The ivga() function has a one more way to call it, which is for internal
% use only:
%
%   result = ivga( data, cs, state, rseed, [cs_args] )
%
% Here a State struct from a previous run can be passed to the controller and
% grouping is then continued from the IVGA state contained in it. The group
% cache and the state transition limiter from the provided State struct will
% used and updated during the run. The first argument, which is normally the
% data, is ignored if a state was given. The returned result and state will
% _not_ be purged before returning them.
%
%
% Implementation details
% ----------------------
%
% Startup and operation:
%
% The ivga() function reads the controller settings from the given function or
% uses preloaded settings. Most command line options are parsed and applied to
% the settings struct with the cshelper_* functions, but some are parsed also
% here in ivga() function. Then the global State struct and STATS object are
% created, the modeler class specified in the settings is initialized, and the
% correct controller is instantiated and started with the specified parameters.
%   When started, the controller is given the controller settings and it may
% access the global variables 'State' and 'STATS'. The State will always
% contain modeler settings and the data. It may also have some active
% groups. STATS will always be newly created and empty. The modeler class is
% in initialized state when the controller starts.
%   When the controller has finished, the global STATS object is stored into
% the global State struct, which in turn is returned by this function.
%
%
% The controller:
%
% The controller must have a run method of type 'this = run( this, settings )',
% which is used to activate it. The controller's inputs are the settings struct
% ('run' method's argument), the State (global) and STATS (global). The
% controller's outputs are the State (global) and STATS (global).
%
% The controller can modify the State directly or with the following functions:
%   IVGA primitive operations:
%     Reinit_model     tries to improve a model by reinitializing it
%     Improve_model    tries to improve a model by continuing adaptation
%     Move_var         tries to move a single var from group to another
%     Merge            tries to merge two groups
%     Split            tries to regroup one or two groups
%   State helper functions:
%     find_group         returns the active group index (State.Ind(i)) of the
%                        group containing the given variable.
%     find_or_create     returns the cache index (State.Groups(i)) of a group
%                        defined by its variables. if no such group exists in
%                        the cache, then one is created and its index is
%                        returned.
%     set_group_active   activates or deactivates a group. its model is
%                        reconstructed/purged. 
% Modeler interface is defined in mdl_definition.txt.
%
% The controller can monitor the situation by directly inspecting the State,
% or by reading from STATS (directly or through some filters/triggers). More
% information about STATS can be found from the stats constructor.
%
% The ctrl_basic controller and its default settings (cs_basic.m) can be used
% as a controller example. The ctrl_adaptive controller demonstrates how to
% inherit from another controller. The ctrl_incremental controller (removed
% from the current version, refer to version ivga-1_10_6 and prior) can be used
% as an example about how to combine multiple controllers.
%
%
% Input:
%
% Input sources if no state struct is provided as function argument:
%   -controller settings: from function argument 'cs', field 'ctrl'.
%   -modeler settings:    from function argument 'cs', field 'mdl'.
%   -data:                from function argument 'data'.
%   -active grouping:     an empty grouping is created.
% Input sources if a state struct is provided as function argument:
%   -controller settings: from function argument 'cs', field 'ctrl'.
%   -modeler settings:    from the given state
%   -data:                from the given state
%   -active grouping:     from the given state
%
%
% Global variables available for the controller:
%
% Global variables 'STATS' and 'State' are created and are available for
% the controller:
%   STATS   progress can be recorded into this 'stats' object.
%   State   current grouping. this is either an empty, newly created state
%           or from a state given by the caller. for internal structure
%           details, see helper function 'make_empty_state'.
% A new stats object is created for the controller, and it is stored into
% the state when the controller finishes. The controller will _always_ get an
% empty, newly created stats object, even if this function is called
% recursively from another controller. In that case the caller's stats will
% be put aside and restored finally.
%
%
% Settings struct/function:
%
% The settings can be prodived as a ready cs struct or by giving a function
% handle. The function should be of type: cs = f( data, state, varargin ),
% where:
%   data       data from the given state or from command line, precedence in
%              this order.
%   state      is either the state given from command line, or [] if none was
%              given.
%   varargin   all extra variables given from command line.
%              note: the function should tolerate unknown arguments!
% The returned settings struct (cs) should should have the following fields:
%   .ctrl   settings for the controller
%   .mdl    settings for the modeler
%           (this is ignored if a state struct was provided)
%   .stl    the state transition limiter to be used
%
% Settings fields required directly in this function:
%   .mdl.class    Class name of the modeler (mdl_*).
%                 (this is ignored if a state struct was provided)
%                 Rest fields (.mdl.*) are for the modeler.
%   .ctrl.class   Class name of the controller (ctrl_*).
%                 Rest fields (.ctrl.*) are for the controller.
%   .stl          The state transition limiter to be used.


% $Id: ivga.m,v 1.59 2006-11-03 13:44:15 pwagner Exp $
% $Name: ivga-1_14_8 $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% set constants (these are read directly from the helper functions with evalin)
DEFAULT_CS              = @cs_basic;
IVGA_VERSION            = '$Name: ivga-1_14_8 $';
RESULT_STRUCT_VERSION   = 3;




% ---- init and configure ----


% check and parse command line arguments
config = configure( data, varargin{:} );

% check the data and create or merge global settings
if ~config.internal_call
  
  % called directly by user
  
  % make sure that the data is from ivga_import_data
  if ~isarg( config.cs_args, 'continue' )
    data = ivga_import_data( data );
  end
  
  % create the global config for this run
  create_global_config( config );
  
else
  
  % recursive call, just merge the current global config into the local config
  config = merge_global_config( config );
  
end




% ---- detect batch mode and handle it separately ----


% activate batch mode if multiple data sets and 'runs' not specified
if iscell( data ) && ~isarg( config.cs_args, 'runs' )
  config.cs_args = { config.cs_args, 'runs', 1 };
end

% handle separately if 'runs' or 'continue' option is specified
if isarg( config.cs_args, 'runs' ) || isarg( config.cs_args, 'continue' )
  result = batch_run_wrapper( data, config );
  return;
end




% ---- initialization ----


% init the random number generator (note that the randn() generator is not
% initialized as it is not currently used!)
rand( 'state', config.rseed );

% create and initialize the global STATS object
global STATS; stats_prev = STATS;
STATS = stats;
STATS = set_active( STATS, 1 );
STATS = set_statevar( STATS, 'internal_quiet', config.quiet );
STATS = add_primitive_ops( STATS );
STATS = cputimer( STATS, 'start' );

% init cs: if a cs function handle was given, get the cs struct from it
cs = config.cs;
if isa( cs, 'function_handle' )
  if isempty(config.state), d = data; else d = config.state.data; end
  cs = feval( cs, d, config.state, config.cs_args{:} );
end

% init state: if no state given, create an empty one
state = config.state;
if isempty( state )
  state = make_empty_state( data, config.rseed, cs, cs.mdl.class, cs.mdl );
end

% globalize state
global State;
state_old = State; State = state;

% register the modeler operations to stats. modeler initialization.
eval([ cs.mdl.class '( ''init'', cs.mdl );' ]);




% ---- computation ----


% instantiate and start the controller and gather stats from its execution.
controller_class = cs.ctrl.class;
if ~config.quiet
  fprintf( 'Controller ''%s'' running:\n', controller_class );
end
STATS = add_operation( STATS, controller_class );
controller = make_controller( controller_class );
STATS = measure( STATS, controller_class, 'start', State );
run( controller, cs.ctrl );
STATS = measure( STATS, controller_class, 'checkpoint', State );
STATS = measure( STATS, controller_class, 'stop', State );
if ~config.quiet && environment( 'has_feature', 'addressable_cursor' )
  fprintf([ 27 '[1A' 27 '[K' ]);
end




% ---- finalization ----


% print a summary and remind the user if there were some problems
if ~config.internal_call, print_summary_and_warnings( config, State ); end

% unglobalize State
state = State;
State = state_old; clear state_old;

% insert the STATS object into state
state.stats = STATS;

% all done, clean up global environment and return the result
STATS       = stats_prev;
if ~config.internal_call
  state = purge_state( state );
  clear global IVGA_CONFIG;
end
result = generate_result( state );

return;








function config = configure( data, varargin )

% Fields of the returned config struct:
%   .cs                   settings, either as a function handle or as a struct
%   .cs_args              contains all the rest arguments from the command line
%   .state                either [] or the final state from a previous run
%   .rseed                the final random seed or state vector to be used
%   .internal_call        true | false
%   .quiet                true | false

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% initialize the configuration with default values
config.cs                 = evalin( 'caller', 'DEFAULT_CS' );
config.cs_args            = {};
config.state              = [];
config.rseed              = 0;
config.internal_call      = false;
config.quiet              = false;


% check the quiet flag
config.quiet = isarg( 'quiet' );

% if the call is a batch 'continue' call, then put all input arguments directly
% into cs_args and return
if isarg( {data, varargin{:}}, 'continue' )
  config.cs_args = {data, varargin{:}};
  return;
end


% detect the call syntax and update the configuration accordingly
%   (cs_args are parsed with isarg and thus always begin with a string
%   argument, and this fact is used in the internal call syntax check)
if nargin == 0
  % ivga()
  error([ 'At least the data argument must be provided! Example: ' ...
          'result = ivga( D ), ' ...
          'where D is a real-valued matrix.' ]);
end
if length(varargin) >= 3 && ...
      ( isstruct(varargin{2}) || isempty(varargin{2}) ) && ...
      ( isnumeric(varargin{3}) && ~isempty(varargin{3}) )
  % ivga( data, cs, state, rseed, [cs_args] )   (the internal syntax)
  config.cs = varargin{1};
  if ~isempty(varargin{2}), config.state = varargin{2}; end
  config.rseed = varargin{3};
  varargin(1:3) = [];
  config.cs_args = varargin;
  config.internal_call = true;
else
  % ivga( data, [rseed], [cs, [cs_args]] )
  if length(varargin) >= 1 && ...
        isnumeric(varargin{1}) && ~isempty(varargin{1})
    % ivga( data, rseed, [cs, [cs_args]] )
    config.rseed = varargin{1};
    varargin(1) = [];
  end
  % ivga( data, [cs], [cs_args] )
  if length(varargin) >= 1 && ...
        ( isa(varargin{1}, 'function_handle') || isstruct(varargin{1}) )
    % ivga( data, cs, [cs_args] )
    config.cs = varargin{1};
    varargin(1) = [];
  end
  % ivga( data, [cs_args] )
  config.cs_args = varargin;
end

% check the validity of the configuration
if ~isa( config.cs, 'function_handle' ) && ~isstruct( config.cs )
  error([ 'The ''cs'' argument must be either a settings function ' ...
          'handle or a preloaded settings struct!' ]);
end
if ~isempty( config.state ) && ~isstruct( config.state )
  error([ 'The ''state'' argument must be either [] or a state struct!' ]);
end
if ~isnumeric( config.rseed ) || ...
      ( length(config.rseed) ~= 1 && length(config.rseed) ~= 35 )
  error([ 'The ''random_seed'' argument must be a real-valued number ' ...
          'or a 35-element vector!' ]);
end

% process the random seed
if config.rseed == 0
  config.rseed = sum(100*clock);
elseif config.rseed == -1
  config.rseed = rand('state');
end

return;




function create_global_config( config )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global IVGA_CONFIG;


NEW_IVGA_CONFIG.VALID           = 'dummy';
NEW_IVGA_CONFIG.QUIET           = config.quiet;
NEW_IVGA_CONFIG.PROPAGATED_ARGS = {};
IVGA_CONFIG = NEW_IVGA_CONFIG;

return;




function config = merge_global_config( config )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global IVGA_CONFIG;


% check that the global config really exists. create if not.
if ~isfield( IVGA_CONFIG, 'VALID' )
  create_global_config( config );
  return;
end

% merge global config into local config
config.quiet   = config.quiet || IVGA_CONFIG.QUIET;
config.cs_args = [ config.cs_args, IVGA_CONFIG.PROPAGATED_ARGS ];

return;




function STATS = add_primitive_ops( STATS )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


STATS = add_operation( STATS, 'Reinit_model', { 'stop trigger level' } );
STATS = add_operation( STATS, 'Improve_model', { 'stop trigger level' } );
STATS = add_operation( STATS, 'Move_var' );
STATS = add_operation( STATS, 'Merge' );
STATS = add_operation( STATS, 'Split', { 'target group count' }, ...
                       'print', {'target group count'} );

STATS = add_operation( STATS, ...
                       'Move_var - Reinit_model', { 'stop trigger level' } );
STATS = add_operation( STATS, ...
                       'Move_var - Improve_model', { 'stop trigger level' } );
STATS = add_operation( STATS, ...
                       'Merge - Reinit_model', { 'stop trigger level' } );
STATS = add_operation( STATS, ...
                       'Merge - Improve_model', { 'stop trigger level' } );

return;




function State = make_empty_state( data, rseed, cs, ...
                                   modeler_class, modeler_settings )

% State = make_empty_state( data, rseed, cs, modeler_class, modeler_settings )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

State.Groups = struct( 'Vars', {}, 'Model', {} );   % group cache
State.Hash   = [];                                  % lookup table for cache
State.Ind    = [];                                  % active group cache inds
State.vars   = [];                                  % active vars

State.state_transition_limiter = cs.stl;   % the state transition limiter

State.data  = data;    % the data
State.rseed = rseed;   % random seed that was used
State.cs    = cs;      % initial settings struct that was used

State.modeler_class    = modeler_class;      % modeler class name
State.modeler_settings = modeler_settings;   % modeler settings struct

return;




function controller = make_controller( cname )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


controller = eval([ cname ';' ]);

return;




function result = generate_result( state )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


result.group_count      = length( state.Ind );
result.grouping         = { state.Groups( state.Ind ).Vars };
for i=1:length(state.Ind)
  result.group_models(i) = ...
      get_group_model_struct( state.Groups( state.Ind(i) ).Model );
end
result.cost             = total_cost( state );
result.random_seed      = state.rseed;
result.settings         = state.cs;
result.internal         = state;

result.internal.ivga_version = ...
    evalin( 'caller', 'IVGA_VERSION' );
result.internal.result_struct_version = ...
    evalin( 'caller', 'RESULT_STRUCT_VERSION' );

return;




function print_summary_and_warnings( config, state )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
global IVGA_CONFIG;


if ~config.quiet
  
  fprintf('\n\n\n\nComputation finished.\n');
  
  % warn about a missing mex file
  if isfield( IVGA_CONFIG, 'MDL_VQ_RN_NONMEX_IN_USE' ) && ...
        IVGA_CONFIG.MDL_VQ_RN_NONMEX_IN_USE
    fprintf(['\n\n*** The C implementation of the Gaussian mixture model ' ...
             'is not compiled for\n    this platform! The slower Matlab ' ...
             'implementation was used instead. You can\n    compile the C ' ...
             'implementation for the current platform with the ' ...
             'command\n    ''ivga_compile''.\n' ...
             '    See the README file for more details.\n\n']);
    warning( 'Failed to use the C implementation of the modeler!' );
  end

  % warn about stability problems
  if isfield( IVGA_CONFIG, 'MDL_VQ_STABILITY_PROBLEMS' ) && ...
        IVGA_CONFIG.MDL_VQ_STABILITY_PROBLEMS > 0
    fprintf(['\n\n*** Some stability problems were detected in the ' ...
             'Gaussian mixture model!\n    Problems were encountered %d ' ...
             'times.\n    You can use the option ''implicit_noisevar'' to ' ...
             'try to stabilize the model.\n    For more details, type: ' ...
             '''help ivga''\n\n'], IVGA_CONFIG.MDL_VQ_STABILITY_PROBLEMS );
    warning( 'Stability problems detected in the modeler!' );
  end
  
  fprintf([ '\nThe grouping can be read directly from the field ' ...
            '''grouping'' in the returned\nstruct. To list the variables ' ...
            'in all groups:\n' ...
            '  <result>.grouping{:}\n' ]);
  fprintf([ '\nYou can also use the functions ivga_show() and ' ...
            'ivga_show_stats() to work with\n' ...
            'the returned result:\n' ...
            '  ivga_show( <result>, [options] )\n' ...
            '  ivga_show_stats( <result>, [options] )\n\n' ] );
  
end


return;








function result = batch_run_wrapper( data, config )

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if isarg( config.cs_args, 'runs' )
  
  % start a new batch
  
  % read the number of iterations
  [flg, runs] = isarg( config.cs_args, 'runs' );
  
  % read the batch id
  [flg, id] = isarg( config.cs_args, 'save_as' ); if ~flg, id = []; end
  
  % remove the 'runs' and 'save_as' arguments from the cs_args
  for i=length(config.cs_args)-1:-1:1
    if ischar(config.cs_args{i}) && ...
          any( strcmp( config.cs_args{i}, {'runs', 'save_as'} ) )
      config.cs_args(i:i+1) = [];
    end
  end
  
  % call batch_run()
  [result, batch_info] = batch_run( id, data, runs, ...
                                    {config.cs}, {config.cs_args}, ...
                                    '', 'random_seed', config.rseed );
  
  % return the result and discard batch_info
  return;
  
else
  
  % continue an interrupted batch
  
  % read the batch id
  [flg, id] = isarg( {data, config.cs_args{:}}, 'continue' );
  if ~flg, error('Internal error'); end
  
  % call batch_run()
  [result, batch_info] = batch_run( id );
  
  % return the result and discard batch_info
  return;
  
end
